/*
 * Linux Ethernet RSTP (IEEE 802.1w) bridging implementation
 * Copyright (C) 2003-2004 IMMS gGmbH, <http://www.imms.de/>
 *
 * Authors:
 * Marco Goetze, <marco.goetze@imms.de>
 *
 * This code builds upon parts of the original STP bridging implementation by
 * Lennert Buytenhek, <buytenh@gnu.org>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include "br_rstp.h"
#include "br_rstp_stm.h"

#define JIFFIES_TO_TICKS(n) (((n) << 8) / HZ)
#define TICKS_TO_JIFFIES(n) (((n) * HZ) >> 8)

void
br_dump_bridge_id(bridge_id *id)
{
	printk("%.2x%.2x.%.2x:%.2x:%.2x:%.2x:%.2x:%.2x",
	       id->prio[0], id->prio[1],
	       id->addr[0], id->addr[1], id->addr[2], id->addr[3],
	       id->addr[4], id->addr[5]);
}

void
br_dump_priority_vector(struct net_bridge_priority_vector *v)
{
	printk("{ ");
	br_dump_bridge_id(&v->RootBridgeID);
	printk(" : %d : ", v->RootPathCost);
	br_dump_bridge_id(&v->DesignatedBridgeID);
	printk(" : %04x : %04x }", v->DesignatedPortID, v->BridgePortID);
}

/* called under ioctl_lock or bridge lock */
struct net_bridge_port*
br_get_port_by_port_no(struct net_bridge *br, int port_no)
{
	struct net_bridge_port *p;

	p = br->port_list;
	while (p != NULL) {
		if (p->port_no == port_no)
			return p;

		p = p->next;
	}

	return NULL;
}

/* called under ioctl_lock or bridge lock */
struct net_bridge_port*
br_get_port_by_portId(struct net_bridge *br, port_id portId)
{
	struct net_bridge_port *p;

	p = br->port_list;
	while (p != NULL) {
		if (p->portId == portId)
			return p;

		p = p->next;
	}

	return NULL;
}

int
br_stp_compare_priority_vectors(struct net_bridge_priority_vector *v1,
                                struct net_bridge_priority_vector *v2,
                                int compareBridgePortID)
{
	int res;

	res = memcmp(&v1->RootBridgeID, &v2->RootBridgeID, 8);
	if (res != 0)
		return res;

	if (v1->RootPathCost != v2->RootPathCost)
		return v1->RootPathCost < v2->RootPathCost ? -1 : 1;

	res = memcmp(&v1->DesignatedBridgeID, &v2->DesignatedBridgeID, 8);
	if (res != 0)
		return res;

	if (v1->DesignatedPortID != v2->DesignatedPortID)
		return v1->DesignatedPortID < v2->DesignatedPortID ? -1 : 1;

	if (compareBridgePortID) {
		if (v1->BridgePortID != v2->BridgePortID)
			return v1->BridgePortID < v2->BridgePortID ? -1 : 1;
	}

	return 0;
}

int
br_stp_compare_times(struct net_bridge_times *t1, struct net_bridge_times *t2)
{
	return t1->FwdDelay  != t2->FwdDelay ||
	       t1->HelloTime != t2->HelloTime ||
	       t1->MaxAge    != t2->MaxAge/* ||
	       t1->MessageAge != t2->MessageAge*/;
}

/* called under bridge lock */
void
br_stp_handle_bpdu(struct sk_buff *skb)
{
	static unsigned char header[5] = { 0x42, 0x42, 0x03, 0x00, 0x00 };

	unsigned char *buf;
	struct net_bridge_port *p;

	buf = skb->mac.raw + 14;
	p = skb->dev->br_port;

	/* 9.3.4, validation */
	if (skb->len < 4+3 || memcmp(buf, header, 5) ||
		(!(buf[5] == 0x00 && buf[6] == 0x00 && skb->len >= 35+3) &&  /* config */
		 !(buf[5] == 0x02 && buf[6] == 0x02 && skb->len >= 36+3 &&   /* RST */
		   p->br->ForceVersion >= 2) &&
		 !(buf[5] == 0x00 && buf[6] == 0x80 && skb->len >= 4+3)))    /* TCN */
	{
		kfree_skb(skb);
		return;
	}

	/* 9.1, decoding according to message structure specification */
	memset(&p->bpdu, 0, sizeof(p->bpdu));
	if (buf[6] == 0x00)
		p->bpdu.type = BR_BPDU_TYPE_CFG;
	else if (buf[6] == 0x02)
		p->bpdu.type = BR_BPDU_TYPE_RST;
	else
		p->bpdu.type = BR_BPDU_TYPE_TCN;

	switch (p->bpdu.type) {
		case BR_BPDU_TYPE_RST:
			switch ((buf[7] & 0x0c) >> 2) {
				case 1:  p->bpdu.role = BR_BPDU_ROLE_ROOT;       break;
				case 2:  p->bpdu.role = BR_BPDU_ROLE_ALT_BKP;    break;
				case 3:  p->bpdu.role = BR_BPDU_ROLE_DESIGNATED; break;
				default: p->bpdu.role = BR_BPDU_ROLE_UNKNOWN;
			}

			p->bpdu.proposal   = (buf[7] & 0x02) ? TRUE : FALSE;
			p->bpdu.learning   = (buf[7] & 0x10) ? TRUE : FALSE;
			p->bpdu.forwarding = (buf[7] & 0x20) ? TRUE : FALSE;
			p->bpdu.agreement  = (buf[7] & 0x40) ? TRUE : FALSE;

			/* fall through */

		case BR_BPDU_TYPE_CFG:
			p->bpdu.tc    = (buf[7] & 0x01) ? TRUE : FALSE;
			p->bpdu.tcAck = (buf[7] & 0x80) ? TRUE : FALSE;

			memcpy(&p->bpdu.priority.RootBridgeID, &buf[8], 8);
			p->bpdu.priority.RootPathCost =
				(buf[16] << 24) | (buf[17] << 16) | (buf[18] << 8) | buf[19];
			memcpy(&p->bpdu.priority.DesignatedBridgeID, &buf[20], 8);
			p->bpdu.priority.DesignatedPortID = (buf[28] << 8) | buf[29];
			p->bpdu.priority.BridgePortID = p->portId;

			p->bpdu.times.MessageAge = TICKS_TO_JIFFIES((buf[30] << 8) | buf[31]);
			p->bpdu.times.MaxAge     = TICKS_TO_JIFFIES((buf[32] << 8) | buf[33]);
			p->bpdu.times.HelloTime  = TICKS_TO_JIFFIES((buf[34] << 8) | buf[35]);
			p->bpdu.times.FwdDelay   = TICKS_TO_JIFFIES((buf[36] << 8) | buf[37]);

			break;

		default:
			break;
	}

	/* 18.2 (IEEE 802.1t), Bridge Detection STM, state BPDU_SEEN */
	if (p->operEdge) {
		printk(KERN_INFO "%s: port %d (%s): no longer edge\n",
			   p->br->dev.name, p->port_no, p->dev->name);
		p->operEdge = FALSE;
	}

	p->rcvdBPDU = TRUE;  /* 17.18.21 */

	/*
	 * Explicitly step the state machines in order to give them a chance to
	 * process the message before it may be overwritten by a subsequent one.
	 */
	br_rstp_stm_step(p->br);

	kfree_skb(skb);
	return;
}

/* called under bridge lock */
void
br_stp_send_bpdu(struct net_bridge_port *p, struct net_bridge_bpdu *bpdu)
{
	unsigned char buf[39];
	__u16 ticks;

	/* 9.1, encoding according to message structure specification */

	memset(buf, 0, sizeof(buf));
	buf[0] = 0x42;
	buf[1] = 0x42;
	buf[2] = 0x03;

	switch (bpdu->type) {
		case BR_BPDU_TYPE_RST:
			buf[5] = 0x02;
			buf[6] = 0x02;

			switch (bpdu->role) {
				case BR_BPDU_ROLE_ROOT:       buf[7] |= 1 << 2; break;
				case BR_BPDU_ROLE_ALT_BKP:    buf[7] |= 2 << 2; break;
				case BR_BPDU_ROLE_DESIGNATED: buf[7] |= 3 << 2; break;
				default: break;
			}

			if (bpdu->proposal)   buf[7] |= 0x02;
			if (bpdu->learning)   buf[7] |= 0x10;
			if (bpdu->forwarding) buf[7] |= 0x20;
			if (bpdu->agreement)  buf[7] |= 0x40;

			/* fall through */

		case BR_BPDU_TYPE_CFG:
			if (bpdu->tc)    buf[7] |= 0x01;
			if (bpdu->tcAck) buf[7] |= 0x80;

			memcpy(&buf[8], &bpdu->priority.RootBridgeID, 8);
			buf[16] = (bpdu->priority.RootPathCost >> 24) & 0xff;
			buf[17] = (bpdu->priority.RootPathCost >> 16) & 0xff;
			buf[18] = (bpdu->priority.RootPathCost >>  8) & 0xff;
			buf[19] = bpdu->priority.RootPathCost         & 0xff;
			memcpy(&buf[20], &bpdu->priority.DesignatedBridgeID, 8);
			buf[28] = (bpdu->priority.DesignatedPortID >> 8) & 0xff;
			buf[29] = bpdu->priority.DesignatedPortID        & 0xff;

			ticks = JIFFIES_TO_TICKS(bpdu->times.MessageAge);
			buf[30] = (ticks >> 8) & 0xff;
			buf[31] = (ticks)      & 0xff;
			ticks = JIFFIES_TO_TICKS(bpdu->times.MaxAge);
			buf[32] = (ticks >> 8) & 0xff;
			buf[33] = (ticks)      & 0xff;
			ticks = JIFFIES_TO_TICKS(bpdu->times.HelloTime);
			buf[34] = (ticks >> 8) & 0xff;
			buf[35] = (ticks)      & 0xff;
			ticks = JIFFIES_TO_TICKS(bpdu->times.FwdDelay);
			buf[36] = (ticks >> 8) & 0xff;
			buf[37] = (ticks)      & 0xff;

			br_stp_transmit_bpdu(p, buf, 38 + (bpdu->type == BR_BPDU_TYPE_RST));

			break;

		case BR_BPDU_TYPE_TCN:
			buf[6] = 0x80;
			br_stp_transmit_bpdu(p, buf, 7);
			break;

		default:
			break;
	}
}

void
br_tick(unsigned long __data)
{
	struct net_bridge *br = (struct net_bridge *)__data;

	read_lock(&br->lock);
	br_rstp_stm_step(br);
	read_unlock(&br->lock);

	br->tick.expires = jiffies + BR_RSTP_STM_TICK;
	add_timer(&br->tick);
}
