/*
 * Linux Ethernet RSTP (IEEE 802.1w) bridging implementation
 * Copyright (C) 2003-2004 IMMS gGmbH, <http://www.imms.de/>
 *
 * Authors:
 * Marco Goetze, <marco.goetze@imms.de>
 *
 * This code builds upon parts of the original STP bridging implementation by
 * Lennert Buytenhek, <buytenh@gnu.org>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include "br_rstp.h"
#include "br_rstp_stm.h"

static unsigned char br_mac_zero[6] = { 0, 0, 0, 0, 0, 0 };

static port_id br_make_port_id(struct net_bridge_port*);
static void br_stp_change_bridge_id(struct net_bridge*, unsigned char*);

/* called under bridge lock */
void
br_stp_disable_bridge(struct net_bridge *br)
{
	struct net_bridge_port *p;

	printk(KERN_INFO "%s: bridge disabled\n", br->dev.name);

	p = br->port_list;
	while (p != NULL) {
		p->portEnabled = FALSE;
		p = p->next;
	}

	br_fdb_cleanup(br);

	del_timer(&br->tick);
	br_rstp_stm_step(br);
}

/* called under bridge lock */
void
br_stp_disable_port(struct net_bridge_port *p)
{
	printk(KERN_INFO "%s: port %i (%s) disabled\n",
	       p->br->dev.name, p->port_no, p->dev->name);

	p->portEnabled = FALSE;
}

/* called under bridge lock */
void br_stp_enable_bridge(struct net_bridge *br)
{
	struct net_bridge_port *p;
	struct timer_list *timer = &br->tick;

	printk(KERN_INFO "%s: bridge enabled\n", br->dev.name);

	p = br->port_list;
	while (p != NULL) {
		if (p->dev->flags & IFF_UP)
			p->portEnabled = TRUE;
		p = p->next;
	}

	init_timer(timer);
	timer->data = (unsigned long) br;
	timer->function = br_tick;
	timer->expires = jiffies + BR_RSTP_STM_TICK;
	add_timer(timer);
}

/* called under bridge lock */
void
br_stp_enable_port(struct net_bridge_port *p)
{
	printk(KERN_INFO "%s: port %i (%s) enabled\n",
	       p->br->dev.name, p->port_no, p->dev->name);

	p->portEnabled = TRUE;
}

/* called under ioctl_lock or bridge lock */
struct net_bridge_port*
br_get_port(struct net_bridge *br, int port_no)
{
	struct net_bridge_port *p;

	p = br->port_list;
	while (p != NULL) {
		if (p->port_no == port_no)
			return p;

		p = p->next;
	}

	return NULL;
}

/* called under bridge lock */
void
br_init_bridge(struct net_bridge *br)
{
	br_rstp_stm_init_bridge_states(br);

	/* 17.17.2 */
	br->BridgeIdentifier.prio[0] = 0x80;
	br->BridgeIdentifier.prio[1] = 0x00;
	memset(br->BridgeIdentifier.addr, 0x00, ETH_ALEN);

	/* 17.17.3 */
	memset(&br->BridgePriority, 0, sizeof(br->BridgePriority));
	br->BridgePriority.RootBridgeID       = br->BridgeIdentifier;
	br->BridgePriority.DesignatedBridgeID = br->BridgeIdentifier;

	/* 17.16 */
#ifdef CONFIG_BRIDGE_DEFAULT_STP
	br->ForceVersion = 0;  /* 17.16.1 */
#else
	br->ForceVersion = 2;  /* 17.16.1 */
#endif
	br->MigrateTime = 3*HZ;  /* 17.16.4 */
	br->TxHoldCount  = 3;    /* 17.16.6 */

	/* 17.4.4 */
	/* 17.16.2 + 17.28.2 */
#ifdef CONFIG_BRIDGE_FORWARD_DELAY
	br->BridgeTimes.FwdDelay = (CONFIG_BRIDGE_FORWARD_DELAY)*HZ;
#else
	br->BridgeTimes.FwdDelay = 15*HZ;
#endif
	/* 17.16.3 + 17.28.2 */
#ifdef CONFIG_BRIDGE_HELLO_TIME
	br->BridgeTimes.HelloTime = (CONFIG_BRIDGE_HELLO_TIME)*HZ;
#else
	br->BridgeTimes.HelloTime = 2*HZ;
#endif
	/* 17.28.2 */
#ifdef CONFIG_BRIDGE_MAX_AGE
	br->BridgeTimes.MaxAge = (CONFIG_BRIDGE_MAX_AGE)*HZ;
#else
	br->BridgeTimes.MaxAge = 20*HZ;
#endif
	br->BridgeTimes.MessageAge = 0;

#ifdef CONFIG_BRIDGE_AGEING_TIME
	br->AgeingTime = (CONFIG_BRIDGE_AGEING_TIME)*HZ;
#else
	br->AgeingTime = 300*HZ;
#endif

	br->rootPortId = 0;
	br->rootPriority = br->BridgePriority;
	br->rootTimes = br->BridgeTimes;
}

/* called under bridge lock */
void
br_init_port(struct net_bridge_port *p)
{
	p->portId = br_make_port_id(p);

	br_rstp_stm_init_port_states(p);

	/* 18.2 (IEEE 802.1t), Bridge Detection STM, state INIT */
#ifdef CONFIG_BRIDGE_DEFAULT_EDGE
	p->operEdge = TRUE;
#else
	p->operEdge = FALSE;
#endif

#ifdef CONFIG_BRIDGE_DEFAULT_P2PMAC
	p->operPointToPointMAC = TRUE;
#else
	p->operPointToPointMAC = FALSE;
#endif

	p->designatedPriority = p->br->BridgePriority;
	p->designatedPriority.BridgePortID = p->portId;
	p->designatedTimes = p->br->BridgeTimes;

	p->msgPriority = p->designatedPriority;
	p->msgTimes = p->designatedTimes;

	memset(&p->bpdu, 0, sizeof(p->bpdu));
	p->bpdu.type = BR_BPDU_TYPE_NONE;

#ifdef CONFIG_BRIDGE_CARRIER
	p->upSince = jiffies;
	memset(p->carrierChanges.data, 0, sizeof(p->carrierChanges.data));
	p->carrierChanges.last =
	  CONFIG_BRIDGE_CARRIER_EXTENDED_CHANGES > 0
	    ? CONFIG_BRIDGE_CARRIER_EXTENDED_CHANGES - 1
	    : 0;
	p->carrierOK = p->carrierLowLevelOK = TRUE;
#endif
}

static
__u16
br_make_port_id(struct net_bridge_port *p)
{
	return (p->priority << 8) | p->port_no;
}

/* called under bridge lock */
static
void
br_stp_change_bridge_id(struct net_bridge *br, unsigned char *addr)
{
	unsigned char oldaddr[6];
	struct net_bridge_port *p;

	memcpy(oldaddr, br->BridgeIdentifier.addr, ETH_ALEN);
	memcpy(br->BridgeIdentifier.addr, addr, ETH_ALEN);
	memcpy(br->dev.dev_addr, addr, ETH_ALEN);
	br->BridgePriority.RootBridgeID = br->BridgeIdentifier;
	br->BridgePriority.DesignatedBridgeID = br->BridgeIdentifier;

	printk(KERN_INFO "%s: set bridge ID to ", br->dev.name);
	br_dump_bridge_id(&br->BridgeIdentifier);
	printk("\n");

	p = br->port_list;
	while (p != NULL) {
		struct net_bridge_priority_vector *vs[3] = {
			&p->designatedPriority, &p->portPriority, &p->msgPriority
		};
		int i;

		for (i = 0; i < 3; i++) {
			if (!memcmp(vs[i]->RootBridgeID.addr, oldaddr, ETH_ALEN) ||
				!memcmp(vs[i]->RootBridgeID.addr, br_mac_zero, ETH_ALEN))
			{
				memcpy(vs[i]->RootBridgeID.addr, addr, ETH_ALEN);
			}
			if (!memcmp(vs[i]->DesignatedBridgeID.addr, oldaddr, ETH_ALEN) ||
				!memcmp(vs[i]->DesignatedBridgeID.addr, br_mac_zero, ETH_ALEN))
			{
				memcpy(vs[i]->DesignatedBridgeID.addr, addr, ETH_ALEN);
			}
		}
		p->reselect = TRUE;

		p = p->next;
	}
}

/* called under bridge lock */
void
br_stp_recalculate_bridge_id(struct net_bridge *br)
{
	unsigned char *addr;
	struct net_bridge_port *p;

	addr = br_mac_zero;

	p = br->port_list;
	while (p != NULL) {
		if (addr == br_mac_zero ||
		    memcmp(p->dev->dev_addr, addr, ETH_ALEN) < 0)
			addr = p->dev->dev_addr;

		p = p->next;
	}

	if (memcmp(br->BridgeIdentifier.addr, addr, ETH_ALEN))
		br_stp_change_bridge_id(br, addr);
}

/* called under bridge lock */
void
br_stp_set_bridge_priority(struct net_bridge *br, int newprio)
{
	struct net_bridge_port *p;

	br->BridgeIdentifier.prio[0] = (newprio >> 8) & 0xFF;
	br->BridgeIdentifier.prio[1] = newprio & 0xFF;

	p = br->port_list;
	while (p != NULL) {
		p->reselect = TRUE;
		p = p->next;
	}
}

/* called under bridge lock */
void br_stp_set_port_priority(struct net_bridge_port *p, int newprio)
{
	__u16 new_port_id;

	p->priority = newprio & 0xFF;
	new_port_id = br_make_port_id(p);

	p->portId = new_port_id;
	p->reselect = TRUE;
}

/* called under bridge lock */
void
br_stp_set_path_cost(struct net_bridge_port *p, int path_cost)
{
	p->PortPathCost = path_cost;
	p->reselect = TRUE;
}

/* called under bridge lock */
void
br_stp_transmit_bpdu(struct net_bridge_port *p,
                     unsigned char *data,
                     int length)
{
	struct net_device *dev;
	struct sk_buff *skb;
	int size;

	size = length + 2*ETH_ALEN + 2;
	if (size < 60)
		size = 60;

	dev = p->dev;

	if ((skb = dev_alloc_skb(size)) == NULL) {
		printk(KERN_INFO "br: memory squeeze!\n");
		return;
	}

	skb->dev = dev;
	skb->protocol = htons(ETH_P_802_2);
	skb->mac.raw = skb_put(skb, size);
	memcpy(skb->mac.raw, bridge_ula, ETH_ALEN);
	memcpy(skb->mac.raw+ETH_ALEN, dev->dev_addr, ETH_ALEN);
	skb->mac.raw[2*ETH_ALEN] = 0;
	skb->mac.raw[2*ETH_ALEN+1] = length;
	skb->nh.raw = skb->mac.raw + 2*ETH_ALEN + 2;
	memcpy(skb->nh.raw, data, length);
	memset(skb->nh.raw + length, 0xa5, size - length - 2*ETH_ALEN - 2);

	dev_queue_xmit(skb);
}
