/*
 * Linux Ethernet RSTP (IEEE 802.1w) bridging implementation
 * Copyright (C) 2003-2004 IMMS gGmbH, <http://www.imms.de/>
 *
 * This file implements the Port Information STM of 17.21.
 *
 * Authors:
 * Marco Goetze, <marco.goetze@imms.de>
 *
 * This code builds upon parts of the original STP bridging implementation by
 * Lennert Buytenhek, <buytenh@gnu.org>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/if_bridge.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#include "br_private.h"
#include "br_rstp.h"
#include "br_rstp_stm.h"

static void step(struct net_bridge_port*, int*);
br_rstp_stm_port_function br_rstp_stm_port_information = &step;

static net_bridge_bpdu_msg_type rcvBpdu(struct net_bridge_port*);
static unsigned recordProposed(struct net_bridge_port*);
static void setTcFlags(struct net_bridge_port*);
static void updtBPDUVersion(struct net_bridge_port*);
static void updtRcvdInfoWhile(struct net_bridge_port*);

typedef enum {
	BEGIN = BR_RSTP_STM_STATE_BEGIN,
	DISABLED,
	AGED,
	UPDATE,
	CURRENT,
	RECEIVE,
	SUPERIOR,
	REPEAT,
	AGREEMENT
} states;

static
void
enter(struct net_bridge_port *p, states *state, states target)
{
	switch (target) {
		case DISABLED:
			p->rcvdBPDU = p->rcvdRSTP = p->rcvdSTP = FALSE;
			p->portPriority = p->designatedPriority;
			p->portTimes = p->designatedTimes;
			p->updtInfo = p->proposing = FALSE;
			p->agreed = p->proposed = FALSE;
			br_timer_clear(&p->rcvdInfoWhile);
			p->infoIs = BR_INFO_DISABLED;
			p->reselect = TRUE;
			p->selected = FALSE;
#ifdef CONFIG_BRIDGE_CARRIER
			if (*state != DISABLED && !p->carrierOK)
				p->tc = TRUE;
#endif
			break;

		case AGED:
			if (p->infoIs == BR_INFO_RECEIVED) {
				printk(KERN_INFO "%s: neighbor ", p->br->dev.name);
				br_dump_bridge_id(&p->portPriority.DesignatedBridgeID);
				printk(" lost on port %i (%s)\n", p->port_no, p->dev->name);
			}

			p->infoIs = BR_INFO_AGED;
			p->reselect = TRUE;
			p->selected = FALSE;
			break;

		case UPDATE:
			p->portPriority = p->designatedPriority;
			p->portTimes = p->designatedTimes;
			p->updtInfo = FALSE;
			p->agreed = p->synced = FALSE;
			p->proposed = p->proposing = FALSE;
			p->infoIs = BR_INFO_MINE;
			p->newInfo = TRUE;
			break;

		case CURRENT:
			// nothing to be done
			break;

		case RECEIVE:
			p->rcvdMsg = rcvBpdu(p);
			updtBPDUVersion(p);
			setTcFlags(p);
			p->rcvdBPDU = FALSE;
			break;

		case SUPERIOR:
			p->portPriority = p->msgPriority;
			p->portTimes = p->msgTimes;
			updtRcvdInfoWhile(p);
			p->agreed = p->proposing = p->synced = FALSE;
			p->proposed = recordProposed(p);
			p->infoIs = BR_INFO_RECEIVED;
			p->reselect = TRUE;
			p->selected = FALSE;
			break;

		case REPEAT:
			p->proposed = recordProposed(p);
			updtRcvdInfoWhile(p);
			break;

		case AGREEMENT:
			p->agreed = TRUE;
			p->proposing = FALSE;
			break;

		default:
			printk(KERN_INFO "%s: bad transition target in %s: %d -> %d\n",
			       p->br->dev.name, __FILE__, *state, target);
			return;
	}
	*state = target;
}

static
void
step(struct net_bridge_port *p, int *state_val)
{
	states state = *(states*) state_val;
#define enter(target) enter(p, (states*) state_val, target)
	if (state != BEGIN && !p->portEnabled && p->infoIs != BR_INFO_DISABLED)
		enter(DISABLED);
#ifdef CONFIG_BRIDGE_CARRIER
	else if (state != BEGIN && !p->carrierOK && p->infoIs != BR_INFO_DISABLED)
		enter(DISABLED);
#endif
	else {
		switch (state) {
			case BEGIN:
				enter(DISABLED);
				break;

			case DISABLED:
#ifdef CONFIG_BRIDGE_CARRIER
				if (p->portEnabled && p->carrierOK)
#else
				if (p->portEnabled)
#endif
					enter(AGED);
				else if (p->updtInfo || p->rcvdBPDU)
					enter(DISABLED);
				break;

			case AGED:
				if (p->selected && p->updtInfo)
					enter(UPDATE);
				break;

			case UPDATE:
			case SUPERIOR:
			case REPEAT:
			case AGREEMENT:
				enter(CURRENT);
				break;

			case CURRENT:
				if (p->selected && p->updtInfo)
					enter(UPDATE);
				else if (p->infoIs == BR_INFO_RECEIVED &&
				         br_timer_has_expired(&p->rcvdInfoWhile) &&
				         !p->updtInfo && !p->rcvdBPDU)
				{
					enter(AGED);
				} else if (p->rcvdBPDU && !p->updtInfo)
					enter(RECEIVE);
				break;

			case RECEIVE:
				switch (p->rcvdMsg) {
					case BR_MSG_SUP_DESIG:
//printk("SUPERIOR TO:   ");br_dump_priority_vector(&p->portPriority);printk("\n");
						enter(SUPERIOR);
						break;

					case BR_MSG_REP_DESIG:
//printk("REPEATED:      ");br_dump_priority_vector(&p->portPriority);printk("\n");
						enter(REPEAT);
						break;

					case BR_MSG_CONF_ROOT:
//printk("AGREEMENT\n");
						enter(AGREEMENT);
						break;

					default:
//printk("OTHER\n");
						enter(CURRENT);
				}
				break;

			default:
				printk(KERN_INFO "%s: bad state in %s: %d; resetting to BEGIN\n",
					   p->br->dev.name, __FILE__, state);
				enter(BEGIN);
		}
	}
#undef enter
}

/* 17.19.8 */
static
net_bridge_bpdu_msg_type
rcvBpdu(struct net_bridge_port *p)
{
	if (p->bpdu.type == BR_BPDU_TYPE_CFG ||
	    (p->bpdu.type == BR_BPDU_TYPE_RST && p->br->ForceVersion >= 2))
	{
		int res;

		p->msgPriority = p->bpdu.priority;
		p->msgTimes = p->bpdu.times;
//printk("RECEIVED BPDU: ");br_dump_priority_vector(&p->msgPriority);printk("\n");

		/* do a regular priority vector comparison */
		res = br_stp_compare_priority_vectors(&p->msgPriority, &p->portPriority, FALSE);

		if ((p->bpdu.type == BR_BPDU_TYPE_RST && p->bpdu.role == BR_BPDU_ROLE_DESIGNATED) ||
		    p->bpdu.type == BR_BPDU_TYPE_CFG)
		{
			/* SuperiorDesignatedMsg */
			if (res < 0 ||
			    (!memcmp(&p->msgPriority.DesignatedBridgeID, &p->portPriority.DesignatedBridgeID, 8) &&
			     p->msgPriority.DesignatedPortID == p->portPriority.DesignatedPortID &&
			     (res != 0 || br_stp_compare_times(&p->msgTimes, &p->portTimes))))
			 {
				 return BR_MSG_SUP_DESIG;
			 }

			/* RepeatedDesignatedMsg */
			if (res == 0 && !br_stp_compare_times(&p->msgTimes, &p->portTimes))
				return BR_MSG_REP_DESIG;
		}

		/* ConfirmedRootMsg */
		if (p->operPointToPointMAC &&
		    p->bpdu.type == BR_BPDU_TYPE_RST &&
		    p->bpdu.role == BR_BPDU_ROLE_ROOT &&
		    res == 0 &&
		    p->bpdu.agreement)
		{
			return BR_MSG_CONF_ROOT;
		}
	}

	/* OtherMsg */
	return BR_MSG_OTHER;
}

/* 17.19.9 */
static
unsigned
recordProposed(struct net_bridge_port *p)
{
	return p->operPointToPointMAC &&
	       p->br->ForceVersion >= 2 &&
	       p->bpdu.type == BR_BPDU_TYPE_RST &&
	       p->bpdu.role == BR_BPDU_ROLE_DESIGNATED &&
	       p->bpdu.proposal;
}

/* 17.19.13 */
static
void
setTcFlags(struct net_bridge_port *p)
{
	switch (p->bpdu.type) {
		case BR_BPDU_TYPE_CFG:
		case BR_BPDU_TYPE_RST:
			if (p->bpdu.tc)
				p->rcvdTc = TRUE;
			if (p->bpdu.tcAck)
				p->rcvdTcAck = TRUE;
			break;

		case BR_BPDU_TYPE_TCN:
			p->rcvdTcn = TRUE;
			break;

		default:
			break;
	}
}

/* 17.19.18 */
static
void
updtBPDUVersion(struct net_bridge_port *p)
{
	switch (p->bpdu.type) {
		case BR_BPDU_TYPE_CFG:
		case BR_BPDU_TYPE_TCN:
			p->rcvdSTP = TRUE;
			break;

		case BR_BPDU_TYPE_RST:
			if (p->br->ForceVersion >= 2)
				p->rcvdRSTP = TRUE;
			break;

		default:
			break;
	}
}

/* 17.19.19 */
static
void
updtRcvdInfoWhile(struct net_bridge_port *p)
{
	unsigned long age;
	unsigned long l;

	l = p->portTimes.MaxAge / 16 > 1*HZ ? p->portTimes.MaxAge / 16 : 1*HZ;
	// round to nearest whole second
	age = ((p->portTimes.MessageAge + l + HZ/2) / HZ) * HZ;
	if (age <= p->portTimes.MaxAge) {
		unsigned long a = p->portTimes.MaxAge - age;
		unsigned long b = 3*p->portTimes.HelloTime;
		br_timer_set(&p->rcvdInfoWhile, a < b ? a : b);
	} else
		br_timer_clear(&p->rcvdInfoWhile);
}
