/*
 * Linux Ethernet RSTP (IEEE 802.1w) bridging implementation
 * Copyright (C) 2003-2004 IMMS gGmbH, <http://www.imms.de/>
 *
 * This file implements the Port State Transition STM of 17.24.
 *
 * Authors:
 * Marco Goetze, <marco.goetze@imms.de>
 *
 * This code builds upon parts of the original STP bridging implementation by
 * Lennert Buytenhek, <buytenh@gnu.org>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/if_bridge.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#include "br_private.h"
#include "br_rstp_stm.h"

static void step(struct net_bridge_port*, int*);
br_rstp_stm_port_function br_rstp_stm_port_state_transition = &step;

typedef enum {
	BEGIN = BR_RSTP_STM_STATE_BEGIN,
	DISCARDING,
	LEARNING,
	FORWARDING
} states;

static
void
enter(struct net_bridge_port *p, states *state, states target)
{
	switch (target) {
		case DISCARDING:
			if (*state != target) {
				printk(KERN_INFO "%s: port %d (%s): transitioning to DISCARDING state\n",
				       p->br->dev.name, p->port_no, p->dev->name);
			}
			// 17.19.3, disableLearning() is handled implicitly by the input
			// code
			p->learning = FALSE;
			// 17.19.2, disableForwarding() is handled implicitly by the input
			// code
			p->forwarding = FALSE;
			break;

		case LEARNING:
			if (*state != target) {
				printk(KERN_INFO "%s: port %d (%s): transitioning to LEARNING state\n",
				       p->br->dev.name, p->port_no, p->dev->name);
			}
			// 17.19.5, enableLearning() is handled implicitly by the input
			// code
			p->learning = TRUE;
			break;

		case FORWARDING:
			if (*state != target) {
				printk(KERN_INFO "%s: port %d (%s): transitioning to FORWARDING state\n",
				       p->br->dev.name, p->port_no, p->dev->name);
			}
			p->tc = !p->operEdge;
			// 17.19.4, enableForwarding() is handled implicitly by the input
			// code
			p->forwarding = TRUE;
			break;

		default:
			printk(KERN_INFO "%s: bad transition target in %s: %d -> %d\n",
			       p->br->dev.name, __FILE__, *state, target);
			return;
	}
	*state = target;
}

static
void
step(struct net_bridge_port *p, int *state_val)
{
	states state = *(states*) state_val;
#define enter(target) enter(p, (states*) state_val, target)
	switch (state) {
		case BEGIN:
			enter(DISCARDING);
			break;

		case DISCARDING:
			if (p->learn)
				enter(LEARNING);
			break;

		case LEARNING:
			if (p->forward)
				enter(FORWARDING);
			else if (!p->learn)
				enter(DISCARDING);
			break;

		case FORWARDING:
			if (!p->forward)
				enter(DISCARDING);
			break;

		default:
			printk(KERN_INFO "%s: bad state in %s: %d; resetting to BEGIN\n",
			       p->br->dev.name, __FILE__, state);
			enter(BEGIN);
	}
#undef enter
}
